---------------------------- MODULE PaxosWithoutSMT ------------------------
(***************************************************************************)
(* This is a TLA+ specification of the Paxos Consensus algorithm,          *)
(* described in                                                            *)
(*                                                                         *)
(*  Paxos Made Simple:                                                     *)
(*   http://research.microsoft.com/en-us/um/people/lamport/pubs/pubs.html#paxos-simple *)
(*                                                                         *)
(* and a TLAPS-checked proof of its correctness.                           *)
(*                                                                         *)
(* Original spec and proof by L. Lamport.                                  *)
(* This proof was written for comparing the old backend provers with the   *)
(* SMT backend. The results of the evaluation appear in the article        *)
(*                                                                         *)
(*  Encoding TLA+ into Unsorted and Many-Sorted First-Order Logic          *)
(*                                                                         *)
(* This proof was processed with TLAPS v1.0.25464. [H. Vanzetto, 05/2017]  *)
(***************************************************************************)
EXTENDS Integers

THEOREM Zenon == TRUE (*{ by (zenon 10) }*)
ZenonT(X) == TRUE (*{ by (prover:"zenon"; timeout:@) }*)
Isa == TRUE (*{ by (prover:"isabelle") }*)
IsaT(X) ==  TRUE (*{ by (prover:"isabelle"; timeout:@) }*)
IsaM(X) ==  TRUE (*{ by (prover:"isabelle"; tactic:@) }*)
IsaTM(X,Y) ==  TRUE (*{ by (prover:"isabelle"; timeout:@; tactic:@) }*)
THEOREM SimpleArith == TRUE (*{ by (cooper) }*)
AXIOM NoSetContainsEverything == \A S : \E x : x \notin S
THEOREM SetExtensionality == TRUE (*{ by (isabelle "(auto intro: setEqualI)")}*)

THEOREM SMT == TRUE (*{ by (smt) }*)
SMTT(X) == TRUE (*{ by (prover:"smt"; timeout:@) }*)
THEOREM Z3 == TRUE (*{ by (z3) }*)

-----------------------------------------------------------------------------
CONSTANTS Acceptors, Values, Quorums

ASSUME QuorumAssumption ==
          /\ Quorums \subseteq SUBSET Acceptors
          /\ \A Q1, Q2 \in Quorums : Q1 \cap Q2 # {}

(***************************************************************************)
(* The following lemma is an immediate consequence of the assumption.      *)
(***************************************************************************)
LEMMA QuorumNonEmpty == \A Q \in Quorums : Q # {}
BY Zenon, QuorumAssumption

Ballots == Nat

VARIABLES msgs,      \* The set of messages that have been sent.
          maxBallot, \* maxBallot[a] is the highest-number ballot acceptor a
                     \*   has participated in.
          maxVotedBallot, \* maxVotedBallot[a] is the highest ballot in which
          maxVotedValue   \*   a has voted, and maxVotedValue[a] is the value
                          \*   it voted for in that ballot.

vars == <<msgs, maxBallot, maxVotedBallot, maxVotedValue>>

Send(m) == msgs' = msgs \cup {m}

None == CHOOSE v : v \notin Values

LEMMA NoneNotAValue == None \notin Values
BY Zenon, NoSetContainsEverything DEF None

Init == /\ msgs = {}
        /\ maxBallot      = [a \in Acceptors |-> -1]
        /\ maxVotedBallot = [a \in Acceptors |-> -1]
        /\ maxVotedValue  = [a \in Acceptors |-> None]

(***************************************************************************)
(* Phase 1a: A leader selects a ballot number b and sends a 1a message     *)
(* with ballot b to a majority of acceptors.  It can do this only if it    *)
(* has not already sent a 1a message for ballot b.                         *)
(***************************************************************************)
Phase1a(b) ==
  /\ ~ \E m \in msgs : (m.type = "1a") /\ (m.ballot = b)
  /\ Send([type |-> "1a", ballot |-> b])
  /\ UNCHANGED <<maxVotedBallot, maxBallot, maxVotedValue>>

(***************************************************************************)
(* Phase 1b: If an acceptor receives a 1a message with ballot b greater    *)
(* than that of any 1a message to which it has already responded, then it  *)
(* responds to the request with a promise not to accept any more proposals *)
(* for ballots numbered less than b and with the highest-numbered ballot   *)
(* (if any) for which it has voted for a value and the value it voted for  *)
(* in that ballot.  That promise is made in a 1b message.                  *)
(***************************************************************************)
Phase1b(a) ==
  \E m \in msgs :
     /\ m.type = "1a"
     /\ m.ballot > maxBallot[a]
     /\ Send([type |-> "1b",
              ballot |-> m.ballot,
              maxVotedBallot |-> maxVotedBallot[a],
              maxVotedValue |-> maxVotedValue[a],
              acc |-> a])
     /\ maxBallot' = [maxBallot EXCEPT ![a] = m.ballot]
     /\ UNCHANGED <<maxVotedBallot, maxVotedValue>>

(***************************************************************************)
(* Phase 2a: If the leader receives a response to its 1b message (for      *)
(* ballot b) from a quorum of acceptors, then it sends a 2a message to all *)
(* acceptors for a proposal in ballot b with a value v, where v is the     *)
(* value of the highest-numbered proposal among the responses, or is any   *)
(* value if the responses reported no proposals.  The leader can send only *)
(* one 2a message for any ballot.                                          *)
(***************************************************************************)
Phase2a(b) ==
  /\ ~ \E m \in msgs : (m.type = "2a") /\ (m.ballot = b)
  /\ \E v \in Values :
       /\ \E Q \in Quorums,
             S \in SUBSET {m \in msgs : (m.type = "1b") /\ (m.ballot = b)} :
               /\ \A a \in Q : \E m \in S : m.acc = a
               /\ \/ \A m \in S : m.maxVotedBallot = -1
                  \/ \E c \in 0..(b-1) :
                        /\ \A m \in S : m.maxVotedBallot =< c
                        /\ \E m \in S : /\ m.maxVotedBallot = c
                                        /\ m.maxVotedValue = v
       /\ Send([type |-> "2a", ballot |-> b, val |-> v])
  /\ UNCHANGED <<maxBallot, maxVotedBallot, maxVotedValue>>

(***************************************************************************)
(* Phase 2b: If an acceptor receives a 2a message for a ballot numbered    *)
(* b, it votes for the message's value in ballot b unless it has already   *)
(* responded to a 1a request for a ballot number greater than or equal to  *)
(* b.                                                                      *)
(***************************************************************************)
Phase2b(a) ==
  \E m \in msgs :
    /\ m.type = "2a"
    /\ m.ballot >= maxBallot[a]
    /\ Send([type |-> "2b", ballot |-> m.ballot, val |-> m.val, acc |-> a])
    /\ maxBallot' = [maxBallot EXCEPT ![a] = m.ballot]
    /\ maxVotedBallot' = [maxVotedBallot EXCEPT ![a] = m.ballot]
    /\ maxVotedValue' = [maxVotedValue EXCEPT ![a] = m.val]

Next == \/ \E b \in Ballots : Phase1a(b) \/ Phase2a(b)
        \/ \E a \in Acceptors : Phase1b(a) \/ Phase2b(a)

Spec == Init /\ [][Next]_vars
-----------------------------------------------------------------------------
(***************************************************************************)
(* How a value is chosen:                                                  *)
(*                                                                         *)
(* This spec does not contain any actions in which a value is explicitly   *)
(* chosen (or a chosen value learned).  What it means for a value to be    *)
(* chosen is defined by the operator Chosen, where Chosen(v) means that v  *)
(* has been chosen.  From this definition, it is obvious how a process     *)
(* learns that a value has been chosen from messages of type "2b".         *)
(***************************************************************************)
VotedForIn(a, v, b) == \E m \in msgs : /\ m.type = "2b"
                                       /\ m.val  = v
                                       /\ m.ballot  = b
                                       /\ m.acc  = a

ChosenIn(v, b) == \E Q \in Quorums :
                     \A a \in Q : VotedForIn(a, v, b)

Chosen(v) == \E b \in Ballots : ChosenIn(v, b)

(***************************************************************************)
(* The consistency condition that a consensus algorithm must satisfy is    *)
(* the invariance of the following state predicate Consistency.            *)
(***************************************************************************)
Consistency == \A v1, v2 \in Values : Chosen(v1) /\ Chosen(v2) => (v1 = v2)
-----------------------------------------------------------------------------
(***************************************************************************)
(* This section of the spec defines the invariant Inv.                     *)
(***************************************************************************)
Messages ==
       [type : {"1a"}, ballot : Ballots]
  \cup [type : {"1b"}, ballot : Ballots, maxVotedBallot : Ballots \cup {-1},
          maxVotedValue : Values \cup {None}, acc : Acceptors]
  \cup [type : {"2a"}, ballot : Ballots, val : Values]
  \cup [type : {"2b"}, ballot : Ballots, val : Values, acc : Acceptors]

TypeOK == /\ msgs \in SUBSET Messages
          /\ maxVotedBallot \in [Acceptors -> Ballots \cup {-1}]
          /\ maxBallot      \in [Acceptors -> Ballots \cup {-1}]
          /\ maxVotedValue  \in [Acceptors -> Values \cup {None}]
          /\ \A a \in Acceptors : maxBallot[a] >= maxVotedBallot[a]

(***************************************************************************)
(* WontVoteIn(a, b) is a predicate that implies that a has not voted and   *)
(* never will vote in ballot b.                                            *)
(***************************************************************************)
WontVoteIn(a, b) == /\ \A v \in Values : ~ VotedForIn(a, v, b)
                    /\ maxBallot[a] > b

(***************************************************************************)
(* The predicate SafeAt(v, b) implies that no value other than perhaps v   *)
(* has been or ever will be chosen in any ballot numbered less than b.     *)
(***************************************************************************)
SafeAt(v, b) ==
  \A c \in 0..(b-1) :
    \E Q \in Quorums :
      \A a \in Q : VotedForIn(a, v, c) \/ WontVoteIn(a, c)

MsgInv ==
  \A m \in msgs :
    /\ (m.type = "1b") =>
         /\ m.ballot =< maxBallot[m.acc]
         /\ \/ /\ m.maxVotedValue \in Values
               /\ m.maxVotedBallot \in Ballots
               \* conjunct strengthened 2014/04/02 sm
               /\ VotedForIn(m.acc, m.maxVotedValue, m.maxVotedBallot)
               \* /\ SafeAt(m.maxVotedValue, m.maxVotedBallot)
            \/ /\ m.maxVotedValue = None
               /\ m.maxVotedBallot = -1
         \** conjunct added 2014/03/29 sm
         /\ \A c \in (m.maxVotedBallot+1) .. (m.ballot-1) :
               ~ \E v \in Values : VotedForIn(m.acc, v, c)
    /\ (m.type = "2a") =>
         /\ SafeAt(m.val, m.ballot)
         /\ \A ma \in msgs : (ma.type = "2a") /\ (ma.ballot = m.ballot)
                                => (ma = m)
    /\ (m.type = "2b") =>
         /\ \E ma \in msgs : /\ ma.type = "2a"
                             /\ ma.ballot  = m.ballot
                             /\ ma.val  = m.val
         /\ m.ballot =< maxVotedBallot[m.acc]

(***************************************************************************)
(* The following two lemmas are simple consequences of the definitions.    *)
(***************************************************************************)
LEMMA VotedInv ==
        MsgInv /\ TypeOK =>
            \A a \in Acceptors, v \in Values, b \in Ballots :
                VotedForIn(a, v, b) => SafeAt(v, b) /\ b =< maxVotedBallot[a]
  <1> SUFFICES ASSUME MsgInv /\ TypeOK,
                      NEW a \in Acceptors, NEW v \in Values, NEW b \in Ballots,
                      NEW m \in msgs,
                      /\ m.type = "2b"
                      /\ m.val  = v
                      /\ m.ballot  = b
                      /\ m.acc  = a
               PROVE  SafeAt(v, b) /\ b =< maxVotedBallot[a]
    BY Zenon DEF VotedForIn
  <1> QED
    BY Zenon DEF VotedForIn, MsgInv, Messages, TypeOK

LEMMA VotedOnce ==
        MsgInv =>  \A a1, a2 \in Acceptors, b \in Ballots, v1, v2 \in Values :
                       VotedForIn(a1, v1, b) /\ VotedForIn(a2, v2, b) => (v1 = v2)
  <1> SUFFICES ASSUME MsgInv,
                      NEW a1 \in Acceptors, NEW a2 \in Acceptors,
                      NEW b \in Ballots, NEW v1 \in Values, NEW v2 \in Values,
                      VotedForIn(a1, v1, b), VotedForIn(a2, v2, b)
               PROVE  v1 = v2
    BY Zenon
  <1>1. PICK m \in msgs :
          /\ m.type = "2b"
          /\ m.val = v1
          /\ m.ballot = b
          /\ m.acc = a1
    BY Zenon DEF VotedForIn
  <1>2. PICK m2 \in msgs :
          /\ m2.type = "2b"
          /\ m2.val = v2
          /\ m2.ballot = b
          /\ m2.acc = a2
    BY Zenon DEF VotedForIn
  <1>a. PICK ma \in msgs :
           /\ ma.type = "2a"
           /\ ma.ballot = m.ballot
           /\ ma.val = m.val
    BY <1>1, Zenon DEF MsgInv
  <1>b. PICK mb \in msgs :
           /\ mb.type = "2a"
           /\ mb.ballot = m2.ballot
           /\ mb.val = m2.val
    BY <1>2, Zenon DEF MsgInv
  <1>3. /\ MsgInv!(ma)!2
        /\ MsgInv!(mb)!2
        /\ m.ballot = m2.ballot
    BY <1>1, <1>2, <1>a, <1>b, Zenon DEF MsgInv
  <1> ma = mb
    BY ONLY <1>a, <1>b, <1>3, Zenon DEF MsgInv
  <1> QED
    BY <1>1, <1>2, <1>a, <1>b, Zenon DEF MsgInv

AccInv ==
  \A a \in Acceptors:
    /\ (maxVotedValue[a] = None) <=> (maxVotedBallot[a] = -1)
    /\ maxVotedBallot[a] =< maxBallot[a]
    \* conjunct strengthened corresponding to MsgInv 2014/04/02 sm
    /\ (maxVotedBallot[a] >= 0) =>
          VotedForIn(a, maxVotedValue[a], maxVotedBallot[a])  \* SafeAt(maxVotedValue[a], maxVotedBallot[a])
    \* conjunct added corresponding to MsgInv 2014/03/29 sm
    /\ \A c \in Ballots : c > maxVotedBallot[a] => ~ \E v \in Values : VotedForIn(a, v, c)

(***************************************************************************)
(* Inv is the complete inductive invariant.                                *)
(***************************************************************************)
Inv == TypeOK /\ MsgInv /\ AccInv
-----------------------------------------------------------------------------
(***************************************************************************)
(* Arithmetic lemmas                                                       *)
(***************************************************************************)

LEMMA IntLeRefl == \A x \in Int: x <= x BY ONLY SimpleArith
LEMMA IntGeRefl == \A x \in Int: x >= x BY ONLY SimpleArith
LEMMA MinusOneIsInt == -1 \in Int BY SimpleArith
LEMMA NatIsInt == /\ \A x \in Nat : x \in Int
                  /\ \A x \in Nat \union {-1} : x \in Int
  BY ONLY SimpleArith
LEMMA NatIsGe0 == \A x \in Nat: x >= 0 BY SimpleArith
LEMMA IntLeAssoc1 == \A x,y,z \in Int: x > y /\ y >= z => x >= z BY SimpleArith
LEMMA IntLeAssoc2 == \A x,y,z \in Int: z <= y /\ x > y => z <= x BY SimpleArith
LEMMA IntLeAssoc3 == \A x,y,z \in Int: x > y /\ y > z => x > z BY SimpleArith
LEMMA IntLeAssoc4 == \A x,y,z \in Int: x >= y /\ y > z => x > z BY SimpleArith
LEMMA IntLeAssoc5 == \A x,y,z \in Int: x <= y /\ y <= z => x <= z BY SimpleArith
LEMMA IntLeAssoc6 == \A x,y,z \in Int: x < y /\ y <= z => x < z BY SimpleArith
LEMMA IntLeAssoc7 == \A x,y,z \in Int: x <= y /\ y < z => x < z BY SimpleArith
LEMMA IntGeToLe == \A x,y \in Int: x >= y => y <= x BY SimpleArith
LEMMA IntGtIsNeq == \A x,y \in Int: x > y => x # y BY SimpleArith
LEMMA IntLeLeEq == \A x,y \in Int: x <= y /\ y <= x => x = y BY SimpleArith

LEMMA IntIntervalDef ==
  \A i,j \in Int: \A x \in (i + 1) .. (j - 1): x \in Int /\ x > i /\ x < j
BY SimpleArith

LEMMA PositiveIsNat == \A x \in Int: 0 <= x => x \in Nat BY SimpleArith
LEMMA BallotsIsPositive == \A x \in Nat \cup {-1}: 0 <= x + 1 BY SimpleArith
LEMMA IntGtToLe == \A x,y \in Int: x > y => y + 1 <= x BY SimpleArith
LEMMA IntGtEqLt == \A x,y \in Int: x > y <=> y < x BY SimpleArith
LEMMA AddIsInt == \A x,y \in Int: x + y \in Int BY SimpleArith
LEMMA IntDisj == \A x,y \in Int: x < y \/ y <= x BY SimpleArith

(***************************************************************************)
(* The following lemma shows that (the invariant implies that) the         *)
(* predicate SafeAt(v, b) is stable, meaning that once it becomes true, it *)
(* remains true throughout the rest of the excecution.                     *)
(***************************************************************************)
LEMMA SafeAtStable == Inv /\ Next /\ TypeOK' =>
                          \A v \in Values, b \in Ballots:
                                  SafeAt(v, b) => SafeAt(v, b)'
<1> USE DEFS Ballots
<1>0. SUFFICES ASSUME Inv,
                      TypeOK',
                      Next,
                      NEW v \in Values, NEW b \in Ballots,
                      SafeAt(v, b),
                      NEW c \in (0..(b-1))'
               PROVE  (SafeAt(v, b)!(c))'
  BY Zenon DEF SafeAt
<1>Q. PICK Q \in Quorums : SafeAt(v, b)!(c)!(Q)
  BY <1>0, Zenon DEF SafeAt
<1>c. c \in Nat
  BY SimpleArith
<1>1. ASSUME NEW b_1 \in Ballots, Phase1a(b_1)
      PROVE  (SafeAt(v, b)!(c))'
  <2> DEFINE msg == [type |-> "1a", ballot |-> b_1]
  <2>1. msg.type = "1a"
    BY Zenon
  <2>2. /\ maxBallot' = maxBallot
        /\ msgs' = msgs \cup {[type |-> "1a", ballot |-> b_1]}
    BY <1>1(*, Zenon*) DEF Phase1a, Send, TypeOK
  <2>3. \A a \in Q: VotedForIn(a, v, c) => VotedForIn(a, v, c)'
    BY <2>1, <2>2, Zenon DEFS VotedForIn
  <2>4. \A a \in Q: WontVoteIn(a, c) => WontVoteIn(a, c)'
    BY <2>1, <2>2, Zenon DEFS WontVoteIn, VotedForIn
  <2> \A aa \in Q: VotedForIn(aa, v, c)' \/ WontVoteIn(aa, c)'
    BY ONLY <1>Q, <2>3, <2>4, Zenon
  <2> QED
    BY Zenon
<1>2. ASSUME NEW b_1 \in Ballots, Phase2a(b_1)
      PROVE  (SafeAt(v, b)!(c))'
  <2>v. PICK v_1 \in Values: Phase2a(b_1)!2!(v_1)
    BY <1>2 DEFS Phase2a
  <2> DEFINE msg == [type |-> "2a", ballot |-> b_1, val |-> v_1]
  <2>1. msg.type = "2a"
    BY Zenon
  <2>2a. msgs' = msgs \cup {msg}
    BY <2>v DEF Phase2a, Send, TypeOK
  <2>2b. maxBallot' = maxBallot
    BY <1>2, <2>v DEF Phase2a, Send, TypeOK
  <2>3. \A a \in Q: VotedForIn(a, v, c) => VotedForIn(a, v, c)'
    BY <2>1, <2>2a, Zenon DEFS VotedForIn
  <2>4. \A a \in Q: WontVoteIn(a, c) => WontVoteIn(a, c)'
    BY <2>1, <2>2a, <2>2b, Zenon DEFS WontVoteIn, VotedForIn
  <2> \A aa \in Q: VotedForIn(aa, v, c)' \/ WontVoteIn(aa, c)'
    BY ONLY <1>Q, <2>3, <2>4, Zenon
  <2> QED
    BY Zenon
<1>3. ASSUME NEW a \in Acceptors, NEW m \in msgs, Phase1b(a)!(m)
      PROVE  (SafeAt(v, b)!(c))'
  <2> DEFINE msg == [type |-> "1b", ballot |-> m.ballot,
                     maxVotedBallot |-> maxVotedBallot[a],
                     maxVotedValue |-> maxVotedValue[a], acc |-> a]
  <2>1. \A mm : mm = msg => mm.type # "2b"
    <3> SUFFICES ASSUME NEW mm, mm = msg PROVE  mm.type # "2b"
      BY Zenon
    <3> QED
      BY Zenon
  <2>3. \A acc \in Q: VotedForIn(acc, v, c) => VotedForIn(acc, v, c)'
    BY <1>3, <2>1, Zenon DEFS VotedForIn, Send
  <2>4. \A acc \in Q: WontVoteIn(acc, c) => (WontVoteIn(acc, c)!1)'
    BY <1>3, <2>1, Zenon DEFS WontVoteIn, VotedForIn, Send
  <2>5. \A acc \in Q: WontVoteIn(acc, c) => WontVoteIn(acc, c)!2'
    <3> SUFFICES ASSUME NEW acc \in Q, WontVoteIn(acc, c)
                 PROVE  maxBallot'[acc] > c'
      BY Zenon
    <3> maxBallot' \in [Acceptors -> Ballots \cup {-1}]
      BY <1>0, <1>3 DEF Inv, TypeOK
    <3>a. CASE acc = a
      <4> maxBallot'[acc] = m.ballot
        BY <1>3, <3>a
      <4> m.ballot > c
        <5> m.ballot \in Nat /\ maxBallot[a] \in Nat \cup {-1}
          BY <1>0, <1>3 DEFS Inv, TypeOK, Messages
        <5> QED
          BY <1>3, <3>a, NatIsInt, IntLeAssoc3, <1>c, Zenon DEFS WontVoteIn
      <4> QED
        BY <1>3, <2>1, <3>a, QuorumAssumption, Zenon
    <3>b. CASE acc # a
      <4> maxBallot'[acc] = maxBallot[acc]
        BY <1>3, <3>b, QuorumAssumption
      <4> QED
        BY <1>3, <2>1, Zenon DEFS WontVoteIn, VotedForIn, Send
    <3> QED
      BY <3>a, <3>b
  <2> \A acc \in Q: VotedForIn(acc, v, c)' \/ WontVoteIn(acc, c)'
    BY ONLY <1>Q, <2>3, <2>4, <2>5, Zenon DEFS WontVoteIn
  <2> QED
    BY Zenon
<1>4. ASSUME NEW a \in Acceptors, NEW m \in msgs, Phase2b(a)!(m)
      PROVE  (SafeAt(v, b)!(c))'
  <2>2 \A aa \in Acceptors, bb \in Ballots, vv \in Values :
            VotedForIn(aa, vv, bb) => VotedForIn(aa, vv, bb)'
    BY <1>4 DEF TypeOK, VotedForIn, Send
  <2>3. \A aa \in Acceptors, bb \in Ballots : maxBallot[aa] > bb => maxBallot'[aa] > bb
    <3> maxBallot'[a] = m.ballot
      BY <1>0, <1>4 DEF TypeOK
    <3> ASSUME NEW aa \in Acceptors, NEW bb \in Ballots, maxBallot[aa] > bb
        PROVE maxBallot'[aa] > bb
      <4> m.ballot \in Nat /\ maxBallot[aa] \in Nat \cup {-1}
        BY <1>0, <1>4, Zenon DEF TypeOK, Inv, Messages
      <4>1. CASE a = aa
        BY <1>0, <1>4, <4>1, IntLeAssoc4, NatIsInt, Zenon DEF SafeAt
      <4>2. CASE a # aa
        <5> maxBallot'[aa] = maxBallot[aa]
          BY <1>0, <1>4, <4>2, Zenon DEF TypeOK
        <5> QED
          BY <1>0, <1>4, <4>2, IntLeAssoc4, NatIsInt, Zenon DEF SafeAt
      <4> QED
        BY <4>1, <4>2
    <3> QED
      BY <1>0, <1>4, Zenon DEF TypeOK
  <2>4. ASSUME NEW aa \in Acceptors, NEW bb \in Ballots,
               WontVoteIn(aa, bb), NEW vv \in Values,
               VotedForIn(aa, vv, bb)'
        PROVE FALSE
    <3> DEFINE mm == [type |-> "2b", val |-> vv, ballot |-> bb, acc |-> aa]
    <3> aa = a /\ m.ballot = bb
      <4>1. mm \notin msgs
        BY <2>4 DEF WontVoteIn, VotedForIn
      <4>2. mm \in msgs'
        <5>1. PICK m1 \in msgs' :
                /\ m1.type = "2b"
                /\ m1.val = vv
                /\ m1.ballot = bb
                /\ m1.acc = aa
          BY <2>4 DEF VotedForIn
        <5>. QED
          BY <1>0, <5>1, Zenon DEF TypeOK, Messages
      <4> QED
        BY <1>4, <4>1, <4>2 DEF TypeOK, Send
    <3> /\ \A v_1 \in Values : \neg VotedForIn(a, v_1, m.ballot)
        /\ maxBallot[a]  >  m.ballot
        /\ VotedForIn(a, vv, m.ballot)'
      BY <2>4 DEFS WontVoteIn
    <3> maxBallot[a] > m.ballot /\  m.ballot >= maxBallot[a] => FALSE
      <4>1. maxBallot[a] \in Nat \union {-1} /\ m.ballot \in Nat
        BY <1>0 DEF Inv, TypeOK, Messages, Ballots
      <4> DEFINE x == maxBallot[a]
                 y == m.ballot
      <4>2. SUFFICES ASSUME x \in Nat \union {-1}, y \in Nat, x > y, y >= x PROVE FALSE
        BY <4>1
      <4> HIDE DEF x, y
      <4> QED
        BY ONLY <4>2, SimpleArith
    <3> QED
      BY <1>0, <1>4, Zenon
  <2>5 \A aa \in Acceptors, bb \in Ballots : WontVoteIn(aa, bb) => WontVoteIn(aa, bb)'
    BY <2>3, <2>4 DEF WontVoteIn
  <2> \A acc \in Q: VotedForIn(acc, v, c)' \/ WontVoteIn(acc, c)'
    BY ONLY <1>Q, <2>2, <1>c, <2>5, QuorumAssumption, Zenon
  <2> QED
    BY Zenon
<1>5. QED

(***************************************************************************)
(* Type Invariant                                                          *)
(***************************************************************************)

THEOREM TypeOKInv == Spec => []TypeOK
<1> USE DEF Ballots, TypeOK
<1>1. Init => TypeOK
  <2> USE DEF Init
  <2> SUFFICES ASSUME Init
               PROVE  TypeOK
    BY Zenon DEF TypeOK
  <2>1. msgs \in SUBSET Messages
    BY Zenon DEF Messages
  <2>2. maxVotedBallot \in [Acceptors -> Ballots \cup {-1}]
    BY Zenon
  <2>3. maxBallot \in [Acceptors -> Ballots \cup {-1}]
    BY Zenon
  <2>4. maxVotedValue \in [Acceptors -> Values \cup {None}]
    BY Zenon
  <2>5. \A a \in Acceptors : maxBallot[a] >= maxVotedBallot[a]
    BY Zenon, IntGeRefl, MinusOneIsInt
  <2>6. QED
    BY <2>1, <2>2, <2>3, <2>4, <2>5, Zenon DEF TypeOK
<1>2. TypeOK /\ [Next]_vars => TypeOK'
  <2> SUFFICES ASSUME TypeOK,
                      [Next]_vars
               PROVE  TypeOK'
    BY Zenon
  <2>1. ASSUME NEW b \in Ballots, Phase1a(b)
        PROVE  TypeOK'
    <3>1. (msgs \in SUBSET Messages)'
      <4> [type |-> "1a", ballot |-> b] \in Messages
        BY Zenon DEF Messages
      <4> QED
        BY <2>1, ZenonT(60) DEF Phase1a, Messages, Send
    <3>2. (maxVotedBallot \in [Acceptors -> Ballots \cup {-1}])'
      BY <2>1, Zenon DEF Phase1a, Messages, Send
    <3>3. (maxBallot \in [Acceptors -> Ballots \cup {-1}])'
      BY <2>1, Zenon DEF Phase1a, Messages, Send
    <3>4. (maxVotedValue \in [Acceptors -> Values \cup {None}])'
      BY <2>1, Zenon DEF Phase1a, Messages, Send
    <3>5. (\A a \in Acceptors : maxBallot[a] >= maxVotedBallot[a])'
      BY <2>1, Zenon DEF Phase1a, Messages, Send
    <3>6. QED
      BY <3>1, <3>2, <3>3, <3>4, <3>5, Zenon
  <2>2. ASSUME NEW b \in Ballots, Phase2a(b)
        PROVE  TypeOK'
    <3> SUFFICES ASSUME \neg (\E m \in msgs : m.type  =  "2a"  /\  m.ballot  =  b),
                        NEW v \in Values,
                        Phase2a(b)!2!(v)!1,
                        Send([type |-> "2a", ballot |-> b, val |-> v]),
                        UNCHANGED <<maxBallot, maxVotedBallot, maxVotedValue>>
                 PROVE  TypeOK'
      <4> HIDE DEF TypeOK
      <4> QED
        BY <2>2, Zenon DEF Phase2a
    <3>Q. PICK Q \in Quorums,
               S \in SUBSET {m \in msgs : m.type = "1b" /\ m.ballot = b}:
                    /\ \A a \in Q : \E m \in S : m.acc  =  a
                    /\ \/ \A m \in S : m.maxVotedBallot  =  -1
                       \/ \E c \in 0 .. b  -  1 :
                          /\ \A m \in S : m.maxVotedBallot  \leq  c
                          /\ \E m \in S :
                              /\ m.maxVotedBallot  =  c
                              /\ m.maxVotedValue  =  v
      BY Zenon
    <3> [type |-> "2a", ballot |-> b, val |-> v] \in Messages
      BY Zenon DEF Messages
    <3>1. (msgs \in SUBSET Messages)'
      BY Zenon DEF Send, Phase2a, Messages
    <3>2. (maxVotedBallot \in [Acceptors -> Ballots \cup {-1}])'
      BY Zenon DEF Send, Phase2a, Messages
    <3>3. (maxBallot \in [Acceptors -> Ballots \cup {-1}])'
      BY Zenon DEF Send, Phase2a, Messages
    <3>4. (maxVotedValue \in [Acceptors -> Values \cup {None}])'
      BY Zenon DEF Send, Phase2a, Messages
    <3>5. (\A a \in Acceptors : maxBallot[a] >= maxVotedBallot[a])'
      BY Zenon DEF Send, Phase2a, Messages
    <3>6. QED
      BY <2>2, <3>1, <3>2, <3>3, <3>4, <3>5, Zenon DEF TypeOK
  <2>3. ASSUME NEW a \in Acceptors,
               NEW m \in msgs,
               /\ m.type = "1a"
               /\ m.ballot > maxBallot[a]
               /\ Send([type |-> "1b",
                        ballot |-> m.ballot,
                        maxVotedBallot |-> maxVotedBallot[a],
                        maxVotedValue |-> maxVotedValue[a],
                        acc |-> a])
               /\ maxBallot' = [maxBallot EXCEPT ![a] = m.ballot]
               /\ UNCHANGED <<maxVotedBallot, maxVotedValue>>
        PROVE  TypeOK'
    <3>0. [type |-> "1b",
            ballot |-> m.ballot,
            maxVotedBallot |-> maxVotedBallot[a],
            maxVotedValue |-> maxVotedValue[a],
            acc |-> a] \in Messages
      BY Zenon DEF Messages
    <3>1. (msgs \in SUBSET Messages)'
      BY <2>3, <3>0, Zenon DEF Messages, Send
    <3>2. (maxVotedBallot \in [Acceptors -> Ballots \cup {-1}])'
      BY <2>3, Zenon
    <3>3. (maxBallot      \in [Acceptors -> Ballots \cup {-1}])'
      BY <2>3, Zenon DEF Messages
    <3>4. (maxVotedValue  \in [Acceptors -> Values \cup {None}])'
      BY <2>3, Zenon
    <3>5. ASSUME NEW a_1 \in Acceptors
          PROVE maxBallot'[a_1] >= maxVotedBallot'[a_1]
      <4>1. CASE a = a_1
        <5>1. maxBallot'[a_1] = m.ballot
          BY <2>3, <4>1, Zenon
        <5>2. /\ maxBallot[a] >= maxVotedBallot[a]
              /\ maxVotedBallot' = maxVotedBallot
          BY <2>3, Zenon
        <5>3. m.ballot >= maxVotedBallot[a]
          <6>1. /\ m.ballot \in Nat
                /\ maxBallot[a] \in Nat \union {-1}
                /\ maxVotedBallot[a] \in Nat \union {-1}
            BY Zenon DEFS Messages
          <6> QED
            BY <2>3, <4>1, <5>1, <5>2, <6>1, IntLeAssoc1, NatIsInt, Zenon
        <5> QED
          BY <4>1, <5>2, <5>3, <5>1, Zenon
      <4>2. CASE a # a_1
        BY <2>3, <4>2, Zenon
      <4> QED
        BY <4>1, <4>2
    <3>6. QED
      BY <3>1, <3>2, <3>3, <3>4, <3>5, Zenon DEF TypeOK
  <2>4. ASSUME NEW a \in Acceptors,
               NEW m \in msgs,
               /\ m.type = "2a"
               /\ m.ballot >= maxBallot[a]
               /\ Send([type |-> "2b", ballot |-> m.ballot, val |-> m.val, acc |-> a])
               /\ maxBallot' = [maxBallot EXCEPT ![a] = m.ballot]
               /\ maxVotedBallot' = [maxVotedBallot EXCEPT ![a] = m.ballot]
               /\ maxVotedValue' = [maxVotedValue EXCEPT ![a] = m.val]
        PROVE  TypeOK'
    <3>0. [type |-> "2b", ballot |-> m.ballot, val |-> m.val, acc |-> a] \in Messages
      BY <2>4, Zenon DEF Messages
    <3>1. (msgs \in SUBSET Messages)'
      BY <2>4, <3>0, Zenon DEF Messages, Send
    <3>2. (maxVotedBallot \in [Acceptors -> Ballots \cup {-1}])'
      BY <2>4, Zenon DEF Messages
    <3>3. (maxBallot      \in [Acceptors -> Ballots \cup {-1}])'
      BY <2>4, Zenon DEF Messages
    <3>4. (maxVotedValue  \in [Acceptors -> Values \cup {None}])'
      BY <2>4, Zenon DEF Messages
    <3>5. ASSUME NEW a_1 \in Acceptors PROVE maxBallot'[a_1] >= maxVotedBallot'[a_1]
      <4>1. CASE a = a_1
        <5>1. m.ballot >= m.ballot
          <6> \A x \in Nat \cup {-1}: x >= x
            BY SimpleArith
          <6>1. m.ballot \in Nat \cup {-1}
            BY <2>4, Zenon DEFS Messages
          <6> QED
            BY <6>1
        <5>2. /\ maxBallot'[a_1] = m.ballot
              /\ maxVotedBallot'[a_1] = m.ballot
          BY <2>4, <4>1, Zenon
        <5> QED
          BY <4>1, <5>1, <5>2, Zenon
      <4>2. CASE a # a_1
        BY <2>4, <4>2, Zenon
      <4> QED
        BY <4>1, <4>2
    <3>6. QED
      BY <3>1, <3>2, <3>3, <3>4, <3>5, Zenon DEF TypeOK
  <2>5. CASE UNCHANGED vars
    BY <2>5, Zenon DEF vars
  <2>6. QED
    BY <2>1, <2>2, <2>3, <2>4, <2>5, Zenon DEF Next, Phase1b, Phase2b
<1>3. QED

(***************************************************************************)

THEOREM AcceptorInv == Spec => []AccInv
<1> USE DEF Ballots, AccInv
<1>1. Init => AccInv
  <2> SUFFICES ASSUME Init,
                      NEW a \in Acceptors
               PROVE AccInv!(a)
    BY DEF AccInv
  <2> USE DEF Init
  <2>b. -1 >= 0 => FALSE BY ONLY SimpleArith
  <2>1. (maxVotedValue[a] = None) <=> (maxVotedBallot[a] = -1)
    BY Zenon
  <2>2. maxVotedBallot[a] =< maxBallot[a]
    BY IntLeRefl, MinusOneIsInt, Zenon
  <2>3. (maxVotedBallot[a] >= 0) =>
           VotedForIn(a, maxVotedValue[a], maxVotedBallot[a])
    BY <2>b, Zenon
  <2>4. ASSUME NEW c \in Ballots, c > maxVotedBallot[a]
               PROVE ~ \E v \in Values : VotedForIn(a, v, c)
    BY Zenon DEFS VotedForIn
  <2>5. QED
    BY <2>1, <2>2, <2>3, <2>4
<1>2. TypeOK /\ AccInv /\ [Next]_vars => AccInv'
  <2> SUFFICES ASSUME TypeOK, AccInv,
                      NEW a \in Acceptors,
                      [Next]_vars
               PROVE  AccInv!(a)'
    BY DEF AccInv
  <2>1. ASSUME NEW b \in Ballots, Phase1a(b) PROVE AccInv!(a)'
    <3>1. ((maxVotedValue[a] = None) <=> (maxVotedBallot[a] = -1))'
      BY <2>1, Zenon DEF Phase1a
    <3>2. (maxVotedBallot[a] =< maxBallot[a])'
      BY <2>1, Zenon DEF Phase1a
    <3>3. ((maxVotedBallot[a] >= 0) =>
              VotedForIn(a, maxVotedValue[a], maxVotedBallot[a]))'
      BY <2>1, Zenon DEF Phase1a, VotedForIn, Send
    <3>4. (\A c \in Ballots : c > maxVotedBallot[a] => ~ \E v \in Values : VotedForIn(a, v, c))'
      <4>1. SUFFICES ASSUME NEW c \in Ballots',
                            c > maxVotedBallot'[a]
                     PROVE  ~ \E v \in Values : VotedForIn(a, v, c)'
        BY Zenon
      <4>2. AccInv!(a)!4!(c)
        BY Zenon
      <4>3. [type |-> "1a", ballot |-> b].type # "2b"
        BY Zenon
      <4> QED
        BY ONLY <2>1, <4>1, <4>2, <4>3, Zenon DEF Phase1a, VotedForIn, Send
    <3>5. QED
      BY <3>1, <3>2, <3>3, <3>4
  <2>2. ASSUME NEW b \in Ballots, Phase2a(b) PROVE AccInv!(a)'
    <3> SUFFICES ASSUME Phase2a(b)!1, Phase2a(b)!2, Phase2a(b)!3 PROVE AccInv!(a)'
      BY <2>2 DEF Phase2a
    <3>v. PICK v \in Values: Phase2a(b)!2!(v)
      BY Phase2a(b)!3, Zenon DEF Phase2a
    <3>x. [type |-> "2a", ballot |-> b, val |-> v].type = "2a"
      BY Zenon
    <3>Q. PICK Q \in Quorums,
               S \in SUBSET {m \in msgs : m.type  =  "1b"  /\  m.ballot  =  b} :
            Phase2a(b)!2!(v)!1!(Q,S)
      BY <3>v, Zenon DEF Phase2a
    <3>1. ((maxVotedValue[a] = None) <=> (maxVotedBallot[a] = -1))'
      BY Phase2a(b)!3, Zenon DEF Phase2a
    <3>2. (maxVotedBallot[a] =< maxBallot[a])'
      BY Phase2a(b)!3, Zenon DEF Phase2a
    <3>3. ((maxVotedBallot[a] >= 0) =>
              VotedForIn(a, maxVotedValue[a], maxVotedBallot[a]))'
      BY ONLY Phase2a(b), AccInv!(a)!3, <3>x, Zenon DEF Phase2a, VotedForIn, Send
    <3>4. ASSUME NEW c \in Ballots, c > maxVotedBallot'[a]
          PROVE ~ \E vv \in Values : VotedForIn(a, vv, c)'
      <4>1. Phase2a(b)!2!(v)!2
        BY <3>v, Zenon DEF Phase2a
      <4> QED
        BY ONLY <3>4, <4>1, Phase2a(b)!3, AccInv!(a)!4, <3>x, Zenon DEF Phase2a, VotedForIn, Send
    <3>5. QED
      BY <3>1, <3>2, <3>3, <3>4
  <2>3. ASSUME NEW a_1 \in Acceptors, NEW m \in msgs, Phase1b(a_1)!(m)
        PROVE AccInv!(a)'
    <3>1. ((maxVotedValue[a] = None) <=> (maxVotedBallot[a] = -1))'
      BY <2>3, Zenon DEF Phase1b
    <3>2. (maxVotedBallot[a] =< maxBallot[a])'
      <4> SUFFICES maxVotedBallot[a] <= maxBallot'[a]
        BY <2>3, AccInv!(a)!2, Zenon
      <4>8. CASE a_1 = a
        <5> SUFFICES maxVotedBallot[a] <= m.ballot
          BY TypeOK, <4>8, <2>3, AccInv!(a)!2, Zenon  DEF TypeOK
        <5> \A x,y,z \in Nat \cup {-1}: x > y /\ z <= y => z <= x
          BY ONLY SimpleArith
        <5> /\ maxBallot[a] \in Nat \cup {-1}
            /\ maxVotedBallot[a] \in Nat \cup {-1}
            /\ m.ballot \in Nat \cup {-1}
          BY Zenon DEF TypeOK, Messages
        <5> QED
          BY <4>8, <2>3, AccInv!(a)!2, Zenon DEF Phase1b
      <4>9. CASE a # a_1
        BY <4>9, <2>3, AccInv!(a)!2, TypeOK, Zenon DEF Phase1b, TypeOK
      <4> QED
        BY <4>8, <4>9
    <3>x. [type |-> "1b", ballot |-> m.ballot,
           maxVotedBallot |-> maxVotedBallot[a_1],
           maxVotedValue |-> maxVotedValue[a_1], acc |-> a_1].type # "2b"
      BY Zenon
    <3>3. ((maxVotedBallot[a] >= 0) =>
              VotedForIn(a, maxVotedValue[a], maxVotedBallot[a]))'
      BY <2>3, <3>x, ZenonT(30) DEF Phase1b, VotedForIn, Send
    <3>4. ASSUME NEW c \in Ballots, c > maxVotedBallot'[a]
          PROVE ~ \E v \in Values : VotedForIn(a, v, c)'
      BY ONLY AccInv!(a)!4, <3>4, <3>x, Phase1b(a_1)!(m), Zenon DEF Phase1b, VotedForIn, Send
    <3>5. QED
      BY <3>1, <3>2, <3>3, <3>4
  <2>4. ASSUME NEW a_1 \in Acceptors, NEW m \in msgs, Phase2b(a_1)!(m)
        PROVE AccInv!(a)'
    <3> DEFINE msg == [type |-> "2b", ballot |-> m.ballot, val |-> m.val, acc |-> a_1]
    <3>msg. msg \in msgs'
      BY ONLY <2>4, Zenon DEF Send
    <3>1. ((maxVotedValue[a] = None) <=> (maxVotedBallot[a] = -1))'
      <4>1. CASE a # a_1
        BY <4>1, <2>4, Zenon DEF Phase2b, TypeOK
      <4>2. CASE a = a_1
        <5> m.ballot # -1
          <6>1. ~ (-1 \in Nat)
            BY ONLY SimpleArith
          <6> QED
            BY ONLY TypeOK, <2>4, NoneNotAValue, <6>1, Zenon DEF TypeOK, Messages
        <5> m.val # None
          BY <2>4, NoneNotAValue, Zenon DEF Phase2b, TypeOK, Messages
        <5> QED
          BY <4>2, <2>4, Zenon DEF Phase2b, TypeOK, Messages
      <4> QED
        BY <4>1, <4>2
    <3>2. (maxVotedBallot[a] =< maxBallot[a])'
      <4>1. CASE a # a_1
        BY <4>1, <2>4, Zenon DEF Phase2b, TypeOK
      <4>2. CASE a = a_1
        <5> SUFFICES m.ballot <= m.ballot
          BY <4>1, <2>4, Zenon DEF Phase2b, TypeOK
        <5>1. \A x \in Nat: x <= x
          BY ONLY SimpleArith
        <5> QED
          BY <5>1, Zenon DEF TypeOK, Messages
      <4> QED
        BY <4>1, <4>2
    <3>3. ((maxVotedBallot[a] >= 0) =>
              VotedForIn(a, maxVotedValue[a], maxVotedBallot[a]))'
      <4>m. m.ballot >= 0
        BY NatIsGe0, Zenon DEFS TypeOK, Messages
      <4>a1. \A x \in Int: x > x => FALSE
        BY ONLY SimpleArith
      <4>1. CASE a # a_1
        <5>1. /\ maxBallot'[a] = maxBallot[a]
              /\ maxVotedBallot'[a] = maxVotedBallot[a]
              /\ maxVotedValue'[a] = maxVotedValue[a]
          BY ONLY <2>4, <4>1, TypeOK, Zenon DEF TypeOK
        <5>2. SUFFICES ASSUME maxVotedBallot[a] >= 0
                       PROVE VotedForIn(a, maxVotedValue[a], maxVotedBallot[a])'
          BY <5>1, Zenon DEF VotedForIn
        <5>3. AccInv!(a)!4!(maxVotedBallot[a])
          BY <4>a1, NatIsInt, Zenon DEF TypeOK
        <5>4. SUFFICES VotedForIn(a, maxVotedValue[a], maxVotedBallot[a])
          BY ONLY <5>4, <5>1, <2>4, Zenon DEF Send, VotedForIn
        <5> QED
          BY AccInv!(a)!3, <5>2, Zenon DEF VotedForIn
      <4>2. CASE a = a_1
        <5>1. /\ maxBallot'[a] = m.ballot
              /\ maxVotedBallot'[a] = m.ballot
              /\ maxVotedValue'[a] = m.val
          BY <2>4, <4>2, Zenon DEF TypeOK
        <5> SUFFICES VotedForIn(a, maxVotedValue[a], maxVotedBallot[a])'
          BY <5>1, <4>m, AccInv!(a)!3 DEF TypeOK
        <5>4. SUFFICES \E m_1 \in msgs' :
                          /\ m_1.type  =  "2b"
                          /\ m_1.val  =  m.val
                          /\ m_1.ballot  =  m.ballot
                          /\ m_1.acc  =  a
          BY ONLY <5>1, <5>4 DEF VotedForIn
        <5> SUFFICES ASSUME msg \in msgs'
                     PROVE /\ msg. type  =  "2b"
                           /\ msg. val  =  m.val
                           /\ msg. ballot  =  m.ballot
                           /\ msg. acc  =  a
          BY ONLY <4>2, <3>msg, Zenon DEF Send
        <5> QED
          BY ONLY <4>2, Zenon
      <4> QED
        BY <4>1, <4>2
    <3>4. ASSUME NEW c \in Ballots, c > maxVotedBallot'[a]
          PROVE ~ \E v \in Values : VotedForIn(a, v, c)'
      <4>8. CASE a = a_1
        <5>1. /\ maxBallot'[a] = m.ballot
              /\ maxVotedBallot'[a] = m.ballot
              /\ maxVotedValue'[a] = m.val
          BY <2>4, <4>8, Zenon DEF TypeOK
        <5>2. c > m.ballot
          BY <4>8, <2>4, <3>4, <5>1, Zenon
        <5>b. m.ballot \in Nat
          BY DEF TypeOK, Messages
        <5>3. c > maxVotedBallot[a]
          <6>1. /\ c > m.ballot
                /\ m.ballot >= maxBallot[a]
                /\ maxBallot[a] >= maxVotedBallot[a]
            BY <4>8, <2>4, <3>4, <5>1, Zenon
          <6>2. \A x,y,z \in Int: x > y /\ y >= z => x > z
            BY ONLY SimpleArith
          <6> QED
            BY <6>1, <6>2, <5>b, NatIsInt DEF TypeOK, Messages
        <5>4. ~ (\E v \in Values : VotedForIn(a, v, c))
          BY ONLY AccInv!(a)!4, <5>3
        <5>5. msg.ballot # c
          <6> \A x,y \in Nat: x > y => x # y
            BY ONLY SimpleArith
          <6> QED
            BY <5>2, <5>b, Zenon
        <5> QED
          BY ONLY <4>8, <2>4, <5>1, <5>4, <5>5, Zenon
          DEFS Phase2b, Send, VotedForIn
      <4>9. CASE a # a_1
        <5>1. /\ maxBallot'[a] = maxBallot[a]
              /\ maxVotedBallot'[a] = maxVotedBallot[a]
              /\ maxVotedValue'[a] = maxVotedValue[a]
          BY ONLY <2>4, <4>9, TypeOK, Zenon DEF TypeOK
        <5>2. ~ (\E v \in Values : VotedForIn(a, v, c))
          BY ONLY <4>9, <2>4, AccInv!(a)!4, <3>4, <5>1, Zenon
        <5>3. msg.acc # a
          BY <4>9, Zenon
        <5> QED
          BY ONLY <4>9, <2>4, <5>1, <5>2, <5>3, Zenon DEFS Send, VotedForIn
      <4> QED
        BY <4>8, <4>9
    <3>5. QED
      BY <3>1, <3>2, <3>3, <3>4
  <2>5. CASE UNCHANGED vars
    BY ONLY <2>5, AccInv!(a), Zenon DEFS vars, VotedForIn
  <2>6. QED
    BY <2>1, <2>2, <2>3, <2>4, <2>5 DEF Next, Phase1b, Phase2b
<1>3. QED

(***************************************************************************)

THEOREM MessageInv == Spec => []MsgInv
<1> USE DEF Ballots, Send
<1>1. Init => MsgInv
  BY Zenon DEF Init, MsgInv
<1>2. Inv /\ TypeOK' /\ [Next]_vars => MsgInv'
  <2>0. SUFFICES ASSUME Inv, TypeOK, TypeOK',
                        MsgInv, Next
                 PROVE  MsgInv'
    <3>1. ASSUME MsgInv, UNCHANGED vars, NEW m \in msgs' PROVE MsgInv!(m)'
      <4>1. m.type = "1b" => MsgInv!(m)!1!2'
        <5>1. MsgInv!(m)!1
          BY <3>1 DEF MsgInv, vars
        <5> QED
          BY ONLY <5>1, UNCHANGED vars, Zenon DEF vars, Inv, MsgInv, VotedForIn
      <4>2. ASSUME m.type = "2a" PROVE MsgInv!(m)!2!2'
        <5>1. MsgInv!(m)!2!2
          BY <3>1, <4>2 DEF MsgInv, vars
        <5>2. MsgInv!(m)!2!2!1'
          BY ONLY <5>1, UNCHANGED vars, ZenonT(60)
          DEFS vars, Inv, MsgInv, VotedForIn, SafeAt, WontVoteIn
        <5>3. MsgInv!(m)!2!2!2'
          BY ONLY UNCHANGED vars, <5>1, <4>2, ZenonT(60)
          DEFS vars, Inv, MsgInv, VotedForIn, SafeAt, WontVoteIn
        <5> QED
          BY <5>2, <5>3
      <4>3. m.type = "2b" => MsgInv!(m)!3!2'
        <5>1. MsgInv!(m)!3
          BY <3>1 DEF MsgInv, vars
        <5> QED
          BY ONLY <5>1, UNCHANGED vars, Zenon DEF vars, Inv, MsgInv, VotedForIn
      <4>4. QED
        BY <4>1, <4>2, <4>3
    <3>9. QED
      BY <3>1 DEF Inv, MsgInv
  <2>m. SUFFICES ASSUME MsgInv, NEW m \in msgs'
                 PROVE  MsgInv!(m)'
    BY <2>0, Zenon DEF MsgInv
  <2>1. ASSUME NEW b \in Ballots, Phase1a(b) PROVE  MsgInv!(m)'
    <3> DEFINE msg == [type |-> "1a", ballot |-> b]
    <3>m. msg.type = "1a"
      BY Zenon
    <3>1. ASSUME m.type = "1b" PROVE MsgInv!(m)!1!2'
      <4>1. m \in msgs
        BY <2>0, <2>1, <3>1, <3>m, Zenon DEFS Phase1a
      <4>2. MsgInv!(m)!1!2
        BY <2>0, <3>1, <4>1 DEF MsgInv
      <4>8. VotedForIn(m.acc, m.maxVotedValue, m.maxVotedBallot) =>
            VotedForIn(m.acc, m.maxVotedValue, m.maxVotedBallot)'
        BY <2>1, <3>1, <3>m, <4>1, <4>2, Zenon DEFS Phase1a, VotedForIn
      <4>9. \A c \in m.maxVotedBallot  +  1 .. m.ballot  -  1 :
             \neg (\E v \in Values : VotedForIn(m.acc, v, c)')
        BY <2>1, <3>1, <3>m, <4>1, <4>2, Zenon DEFS Phase1a, VotedForIn
      <4> QED
        BY <2>1, <3>1, <3>m, <4>1, <4>2, <4>8, <4>9, Zenon DEFS Phase1a
    <3>2. ASSUME m.type = "2a" PROVE MsgInv!(m)!2!2'
      <4>1. m \in msgs
        BY <2>0, <2>1, <3>2, <3>m, Zenon DEFS Phase1a
      <4>2. MsgInv!(m)!2!2
        BY <2>0, <3>2, <4>1 DEF MsgInv
      <4>3. m.ballot \in Ballots /\ m.val \in Values
        BY TypeOK, <2>1, <3>m, <3>2, Zenon DEFS Phase1a, TypeOK, Messages
      <4>4. SafeAt(m.val,m.ballot)'
        BY <2>0, <3>2, <4>2, <4>3, SafeAtStable, Zenon
      <4> QED
        BY <2>0, <2>1, <3>m, <3>2, <4>4, Zenon DEFS Phase1a, MsgInv
    <3>3. ASSUME m.type = "2b" PROVE MsgInv!(m)!3!2'
      BY <2>0, <2>1, <3>3, <3>m, Zenon DEFS Phase1a, MsgInv
    <3>4. QED
      BY <3>1, <3>2, <3>3, Zenon
  <2>2. ASSUME NEW b \in Ballots, Phase2a(b) PROVE MsgInv!(m)'
    <3>0. SUFFICES ASSUME Phase2a(b)!1, NEW v \in Values,
                          Phase2a(b)!2!(v), Phase2a(b)!3
                   PROVE  MsgInv!(m)'
      BY <2>2, Zenon DEF Phase2a
    <3> DEFINE msg == [type |-> "2a", ballot |-> b, val |-> v]
    <3>m. msg.type = "2a" /\ msg.ballot = b
      BY Zenon
    <3>1. ASSUME m.type = "1b" PROVE MsgInv!(m)!1!2'
      <4>1. m \in msgs /\ m # msg
        BY <2>0, <3>0, <3>1, <3>m, Zenon
      <4>2. MsgInv!(m)!1!2
        BY <2>0, <3>1, <4>1, Zenon DEF MsgInv
      <4>8. VotedForIn(m.acc, m.maxVotedValue, m.maxVotedBallot) =>
            VotedForIn(m.acc, m.maxVotedValue, m.maxVotedBallot)'
        BY <3>0, <3>1, <3>m, <4>1, <4>2, Zenon DEFS VotedForIn
      <4>9. ASSUME NEW c \in m.maxVotedBallot + 1 .. m.ballot - 1
            PROVE \neg (\E vv \in Values : VotedForIn(m.acc, vv, c)')
        <5>1. Phase2a(b)!2!(v)!2
          BY <3>0
        <5> QED
          BY ONLY Phase2a(b)!1, <5>1, Phase2a(b)!3,
            <3>1, <3>m, <4>1, <4>2, <4>9, Zenon DEFS VotedForIn
      <4> QED
        BY <2>2, <3>1, <3>m, <4>1, <4>2, <4>8, <4>9, Zenon DEFS Phase2a
    <3>2. ASSUME m.type = "2a" PROVE MsgInv!(m)!2!2'
      <4>m. Phase2a(b)!2!(v)!2 \* msgs' = msg \cup {msg}
        BY <3>0, Zenon
      <4>q. PICK Q \in Quorums,
                 S \in SUBSET { m1 \in msgs : m1.type = "1b" /\ m1.ballot = b } :
                 /\ \A a \in Q: \E m0 \in S: m0.acc = a
                 /\ \/ \A m1 \in S : m1.maxVotedBallot = -1
                    \/ \E bmax \in 0 .. b - 1:
                          /\ \A m1 \in S: m1.maxVotedBallot <= bmax
                          /\ \E m1 \in S:
                                /\ m1.maxVotedBallot = bmax
                                /\ m1.maxVotedValue = v
        BY <3>0, Zenon
      <4>2. m.val \in Values /\ m.ballot \in Ballots
        BY TypeOK, <2>1, <3>m, <3>2, <2>0, Zenon DEFS Phase2a, TypeOK, Messages
      <4>arith. ~(-1 \in Nat) /\ -1 + 1 = 0
        BY ONLY SimpleArith
      <4>9. ASSUME NEW n \in msgs', n.type = "2a", n.ballot = m.ballot
            PROVE n = m
        <5>a. CASE m \in msgs
          <6>0. MsgInv!(m)!2!2
            BY <2>m, <3>2, <4>9, <5>a DEF MsgInv
          <6>1. CASE n \in msgs
            BY Phase2a(b)!2!(v), <3>m, <4>9, <5>a, <6>0, <6>1, Zenon
          <6>2. CASE n = msg
            BY Phase2a(b)!1, Phase2a(b)!3, <3>2, <4>9, <5>a, <6>2
          <6> QED
            BY Phase2a(b)!2!(v), <6>1, <6>2
        <5>b. CASE m = msg
          BY <3>m, <4>9, <5>b, <3>2, <3>0, Zenon DEFS Phase2a, MsgInv
        <5> QED
          BY Phase2a(b)!2!(v), <5>a, <5>b, Zenon DEF Phase2a
      <4>8. SafeAt(m.val,m.ballot)'
        <5> SUFFICES SafeAt(m.val,m.ballot)
          <6> \A aa, vv, cc: VotedForIn(aa,vv,cc) <=> VotedForIn(aa,vv,cc)'
            BY <4>m, <3>m, Zenon DEF VotedForIn
          <6> QED
            BY Phase2a(b)!3(*, Zenon*) DEF SafeAt, WontVoteIn
        <5>a. CASE m \in msgs
          BY <2>m, <3>0, <3>2, <5>a, Zenon DEF Phase2a, MsgInv
        <5>b. CASE m = msg
          <6> SUFFICES ASSUME NEW c \in 0 .. b - 1
                       PROVE  \E QQ \in Quorums: \A a \in QQ:
                                VotedForIn(a, v, c) \/ WontVoteIn(a, c)
            BY <5>b, Zenon DEF SafeAt
          <6>c. c \in Int /\ c < b /\ c \in Nat /\ b \in Int
            BY ONLY SimpleArith
          <6>a. CASE \A mm \in S: mm.maxVotedBallot = -1
            <7>1. ASSUME NEW mm \in S
                  PROVE  /\ b =< maxBallot[mm.acc]
                         /\ mm.maxVotedValue = None
                         /\ \A vv \in Values: ~ VotedForIn(mm.acc, vv, c)
              <8> mm.ballot = b
                BY Zenon
              <8> QED
                BY <2>0, <6>a, <4>arith, Zenon DEF MsgInv
            <7> SUFFICES ASSUME NEW a \in Q
                         PROVE VotedForIn(a, v, c) \/ WontVoteIn(a, c)
              BY Zenon
            <7>a. \A vv \in Values: ~ VotedForIn(a, vv, c)
              BY <4>q, <7>1, Zenon
            <7>t. maxBallot[a] \in Nat \cup {-1}
              BY TypeOK, QuorumAssumption, Zenon DEF TypeOK
            <7>b. c < maxBallot[a]
              BY <4>arith, <6>c, <7>t, <7>1, <4>q, IntLeAssoc6, NatIsInt
            <7> QED
              BY Phase2a(b)!1, Phase2a(b)!3, <4>q, <7>a,  <7>b, IntGtEqLt
              DEF MsgInv, WontVoteIn
          <6>b. ASSUME NEW bmax \in 0 .. b - 1,
                       \A m0 \in S: m0.maxVotedBallot <= bmax,
                       NEW mmax \in S,
                       mmax.maxVotedBallot = bmax,
                       mmax.maxVotedValue = v
                PROVE  \E QQ \in Quorums: \A a \in QQ:
                         VotedForIn(a, v, c) \/ WontVoteIn(a, c)
            <7>1. MsgInv!(mmax)!1!2
              BY <2>0, <6>b, Zenon DEF MsgInv
            <7>2. /\ b =< maxBallot[mmax.acc]
                  /\ \A cc \in bmax + 1 .. b - 1:
                       \A vv \in Values: ~ VotedForIn(mmax.acc, vv, cc)
              BY <6>b, <7>1, Zenon DEF MsgInv
            <7> mmax.acc \in Acceptors
              BY TypeOK DEFS TypeOK, Messages
            <7>a. CASE /\ v \in Values
                       /\ bmax \in Nat
                       /\ VotedForIn(mmax.acc, v, bmax)
              <8>a. CASE c < bmax
                <9>1. SafeAt(v,bmax)
                  BY MsgInv, TypeOK, VotedInv, <7>a
                <9>2. c \in 0 .. bmax - 1
                  BY SimpleArith, <8>a
                <9> QED
                  BY <9>1, <9>2, <8>a DEF SafeAt
              <8>b. CASE bmax <= c
                <9> SUFFICES ASSUME NEW a \in Q
                             PROVE  VotedForIn(a, v, c) \/ WontVoteIn(a, c)
                  BY Zenon
                <9>ma. PICK ma \in S: ma.acc = a
                  BY <4>q
                <9>2. /\ b <= maxBallot[a]
                      /\ \A cc \in (ma.maxVotedBallot+1) .. (b-1) :
                           ~ \E vv \in Values : VotedForIn(a, vv, cc)
                  BY <2>0, <9>ma, Zenon DEF MsgInv
                <9>3. c < maxBallot[a]
                  <10>2. b <= maxBallot[a]
                    BY <9>2, <9>ma, Zenon
                  <10>t. maxBallot[a] \in Nat \cup {-1}
                    BY TypeOK, QuorumAssumption, Zenon DEF TypeOK
                  <10> QED
                    BY <6>c, <10>2, <9>ma, IntLeAssoc6, <10>t, NatIsInt, Zenon
                <9>a. CASE /\ ma.maxVotedValue \in Values
                           /\ ma.maxVotedBallot \in Ballots
                           /\ VotedForIn(a, ma.maxVotedValue, ma.maxVotedBallot)
                  <10>1. bmax \in Int /\ ma.maxVotedBallot \in Int
                    BY SimpleArith
                  <10>2. ma.maxVotedBallot \in Int
                    BY <9>a, NatIsInt
                  <10>a. CASE bmax < c
                    <11>3. ma.maxVotedBallot < c
                      BY <6>b, <10>a, <9>a, <10>1, <6>c, <10>2, IntLeAssoc7
                    <11>4. \A x \in Int: x < c => c \in x+1 .. b-1
                      BY SimpleArith
                    <11>5. c \in (ma.maxVotedBallot+1) .. (b-1)
                      BY <10>1, <6>c, <10>2, <11>3, <11>4
                    <11>6. ~ \E vv \in Values : VotedForIn(a, vv, c)
                      BY <9>2, <11>5
                    <11> QED
                      BY <9>3, <11>6 DEF WontVoteIn
                  <10>b. CASE bmax = c
                    <11>1. /\ c \in Int
                           /\ \A x \in Int: x < c => c \in x+1 .. b-1
                      BY ONLY <9>a, SimpleArith
                    <11>a. CASE ma.maxVotedBallot < c
                      <12>2. ma.maxVotedBallot \in Int
                        BY NatIsInt, TypeOK DEF TypeOK, Messages
                      <12> c \in ma.maxVotedBallot + 1 .. b - 1
                        BY ONLY <11>a, <11>1, <12>2
                      <12> QED
                        BY <9>3, IntGtEqLt, <9>2, Zenon DEF WontVoteIn
                    <11>b. CASE c <= ma.maxVotedBallot
                      <12> c = ma.maxVotedBallot
                        BY <11>b, <10>b, <6>b, <10>b, <10>1, <6>c, IntLeLeEq
                      <12>4. VotedForIn(mmax.acc, v, c)
                        BY <7>a, <10>b
                      <12> QED
                        BY MsgInv, <9>a, <12>4, <9>a, VotedOnce, QuorumAssumption, Zenon
                    <11> QED
                      BY <10>b, <8>b, <11>a, <11>b, NatIsInt, <9>a, IntDisj, <6>c
                  <10> QED
                    <11> bmax < c \/ bmax = c
                      BY <8>b, SimpleArith
                    <11> QED
                      BY <10>a, <10>b
                <9>b. CASE /\ ma.maxVotedValue = None
                           /\ ma.maxVotedBallot = -1
                  <10> \neg (\E vv \in Values : VotedForIn(a, vv, c))
                    BY <9>2, <4>arith, <9>b
                  <10> QED
                    BY <9>3 DEF WontVoteIn
                <9> QED
                  BY <9>a, <9>b, <2>0, <9>ma DEF MsgInv
              <8> QED
                BY <8>a, <8>b, IntDisj, <6>c, <7>a, NatIsInt
            <7>b. mmax.maxVotedBallot # -1
              <8> bmax # -1
                BY SimpleArith
              <8> QED
                BY <6>b
            <7> QED
              BY <6>b, <7>1, <7>a, <7>b
          <6> QED
            BY <6>a, <6>b, <4>q
        <5> QED
          BY <4>m, <5>a, <5>b
      <4> QED
        BY <4>8, <4>9
    <3>3. ASSUME m.type = "2b" PROVE MsgInv!(m)!3!2'
      <4>1. m \in msgs /\ m # msg
        BY <2>0, <3>0, <3>3, <3>m, Zenon
      <4>2. MsgInv!(m)!3!2
        BY <2>0, <3>3, <4>1 DEF MsgInv
      <4> QED
        BY <3>0, <3>3, <3>m, <4>2, Zenon DEFS Phase2a
    <3>4. QED
      BY <3>1, <3>2, <3>3
  <2>3. ASSUME NEW a \in Acceptors, NEW mm \in msgs, Phase1b(a)!(mm)
        PROVE  MsgInv!(m)'
    <3> DEFINE msg == [type |-> "1b", ballot |-> mm.ballot,
                       maxVotedBallot |-> maxVotedBallot[a],
                       maxVotedValue |-> maxVotedValue[a],
                       acc |-> a]
    <3>m. msg.type = "1b" /\ msg.acc = a /\ msg.ballot = mm.ballot
      BY Zenon
    <3>1. ASSUME m.type = "1b" PROVE MsgInv!(m)!1!2'
      <4>1. CASE m \in msgs
        <5>0. MsgInv!(m)!1!2
          BY <2>m, <3>1, <4>1, Zenon DEFS MsgInv
        <5>1. m.ballot <= maxBallot'[m.acc]
          <6>1. CASE a = m.acc
            <7>1. /\ m.acc \in Acceptors
                  /\ m.ballot \in Int
                  /\ maxBallot[a] \in Int
                  /\ mm.ballot \in Int
              BY ONLY NatIsInt, <4>1, <6>1, TypeOK, Zenon DEF TypeOK, Messages
            <7>2. maxBallot'[m.acc] = mm.ballot
              BY TypeOK, <2>3, <6>1, <7>1 DEF TypeOK
            <7>3. QED
              BY ONLY IntLeAssoc2, NatIsInt, <2>3, <3>1, <5>0, <6>1, <7>1, <7>2, Zenon
          <6>2. CASE a # m.acc
            <7>1. /\ m.acc \in Acceptors
                  /\ m.ballot \in Int
                  /\ maxBallot[a] \in Int
                  /\ mm.ballot \in Int
              BY ONLY NatIsInt, <3>1, <4>1, <6>2, TypeOK, Zenon DEF TypeOK, Messages
            <7>2. maxBallot'[m.acc] = maxBallot[m.acc]
              BY TypeOK, <2>3, <6>2, <7>1 DEF TypeOK
            <7>3. QED
              BY ONLY <5>0, <7>2, Zenon
          <6> QED
            BY <6>1, <6>2
        <5>8. VotedForIn(m.acc, m.maxVotedValue, m.maxVotedBallot) =>
              VotedForIn(m.acc, m.maxVotedValue, m.maxVotedBallot)'
          BY <2>3, <3>1, <3>m, <4>1, Zenon DEFS Phase1b, VotedForIn
        <5>9. ASSUME NEW c \in m.maxVotedBallot + 1 .. m.ballot - 1
              PROVE \neg (\E v \in Values : VotedForIn(m.acc, v, c)')
          BY ONLY <2>3, <3>1, <3>m, <4>1, <5>0, <5>9, Zenon
          DEFS Phase1b, Send, VotedForIn
        <5> QED
          BY <5>0, <5>1, <5>8, <5>9, Zenon
      <4>2. CASE m = msg
        <5>0. /\ m.ballot \in Int
              /\ maxBallot[a] \in Int
              /\ mm.ballot \in Int
              /\ m.acc = a
          BY ONLY NatIsInt, <4>2, TypeOK, Zenon DEF TypeOK, Messages
        <5>1. m.ballot =< maxBallot'[m.acc]
          <6>1. maxBallot'[a] = mm.ballot
            BY TypeOK, <2>3, <4>2, <5>0 DEF TypeOK
          <6>2. QED
            BY ONLY IntLeRefl, NatIsInt, <4>2, <5>0, <6>1, Zenon
        <5>2. MsgInv!(m)!1!2!2'
          <6>0. /\ m.maxVotedBallot = maxVotedBallot[a]
                /\ m.maxVotedValue = maxVotedValue[a]
            BY <4>2
          <6>1. CASE maxVotedBallot[a] # -1
            <7> /\ m.maxVotedBallot \in Nat
                /\ m.maxVotedValue \in Values
              BY <2>0, <4>2, <6>1, Zenon DEFS Inv, TypeOK, Messages, AccInv
            <7> VotedForIn(m.acc, m.maxVotedValue, m.maxVotedBallot)
              BY <2>0, <2>3, <3>1, <4>2, <6>0, <6>1, NatIsGe0, Zenon
              DEFS Phase1b, Inv, AccInv
            <7> QED
              BY <2>m, <2>3, <3>1, <6>0, <6>1, Zenon
              DEFS Phase1b, VotedForIn, MsgInv
          <6>2. CASE maxVotedBallot[a] = -1
            <7> m.maxVotedValue = None
              BY <2>0, <4>2, <6>2, Zenon DEFS Inv, AccInv
            <7> QED
              BY <2>3, <3>1, <4>2, <6>2, Zenon
              DEFS Phase1b, VotedForIn, MsgInv
          <6> QED
            BY TypeOK, <6>1, <6>2 DEF TypeOK
        <5>3. ASSUME NEW c \in m.maxVotedBallot + 1 .. m.ballot - 1
              PROVE \neg (\E v \in Values : VotedForIn(m.acc, v, c)')
          <6>1. m.maxVotedBallot \in Ballots \cup {-1} /\ m.ballot \in Ballots
            BY <4>2, <3>m, TypeOK DEFS TypeOK, Messages
          <6>2. c \in Ballots
            <7> SUFFICES 0 <= c
              BY <6>1, PositiveIsNat, IntIntervalDef, NatIsInt
            <7>1. 0 <= m.maxVotedBallot + 1
              BY <6>1, BallotsIsPositive
            <7>2. m.maxVotedBallot + 1 <= c
              BY <6>1, PositiveIsNat, IntGtToLe, IntIntervalDef, NatIsInt
            <7>3. m.maxVotedBallot + 1 \in Int
              BY <6>1, NatIsInt, AddIsInt
            <7> QED
              BY <6>1, <7>1, <7>2, <7>3, IntLeAssoc5, IntIntervalDef, NatIsInt, Zenon
          <6>3. c > m.maxVotedBallot
            BY <6>1, IntIntervalDef, NatIsInt
          <6>4. c > maxVotedBallot[a]
            BY <4>2, <6>1, <2>0, <5>3, <6>2, <6>3, Zenon DEF Inv, AccInv
          <6>5. ~ \E v \in Values : VotedForIn(a, v, c)
            BY <2>0, <5>3, <6>2, <6>4, Zenon DEF Inv, AccInv
          <6> QED
            BY <2>3, <3>1, <4>2, <6>5, Zenon DEFS VotedForIn
        <5>4. QED
          BY <5>1, <5>2, <5>3
      <4> QED
        BY <2>3, <4>1, <4>2 DEFS Send
    <3>2. ASSUME m.type = "2a" PROVE MsgInv!(m)!2!2'
      <4>1. m \in msgs
        BY <2>0, <2>3, <3>2, <3>m, Zenon DEFS Phase1b
      <4>2. MsgInv!(m)!2!2
        BY <2>0, <3>2, <4>1 DEF MsgInv
      <4>3. m.ballot \in Ballots /\ m.val \in Values
        BY TypeOK, <2>3, <3>m, <3>2, Zenon DEFS Phase1b, TypeOK, Messages
      <4>4. SafeAt(m.val,m.ballot)'
        BY <2>0, <3>2, <4>2, <4>3, SafeAtStable, Zenon
      <4> QED
        BY <2>0, <2>3, <3>m, <3>2, <4>4, Zenon DEFS Phase1b, MsgInv
    <3>3. ASSUME m.type = "2b" PROVE MsgInv!(m)!3!2'
      BY <2>m, <2>3, <3>m, <3>3, Zenon DEFS Phase1b, Send, MsgInv
    <3>4. QED
      BY <3>1, <3>2, <3>3
  <2>4. ASSUME NEW a \in Acceptors, NEW mm \in msgs, Phase2b(a)!(mm)
        PROVE  MsgInv!(m)'
    <3> DEFINE msg == [type |-> "2b", ballot |-> mm.ballot, val |-> mm.val, acc |-> a]
    <3>m. msg.type = "2b"
      BY Zenon
    <3>1. CASE m.acc = a
      <4>1. maxBallot'[m.acc] = mm.ballot
        BY <2>4, <3>1, TypeOK, Zenon DEFS TypeOK
      <4>2. ASSUME m.type = "1b" PROVE MsgInv!(m)!1!2'
        <5>1. m \in msgs
          BY <2>4, <4>2, <3>m, Zenon DEFS TypeOK, Messages
        <5>2. MsgInv!(m)!1!2
          BY <2>m, <4>2, <5>1, Zenon DEFS MsgInv
        <5>3. /\ m.ballot \in Nat
              /\ mm.ballot \in Nat
              /\ maxBallot[a] \in Nat \cup {-1}
          BY <5>1, TypeOK, Zenon DEFS TypeOK, Messages
        <5>4. m.ballot <= maxBallot'[m.acc]
          BY <2>4, <3>1, <5>2, <4>1, <5>3, IntLeAssoc5, IntGeToLe, NatIsInt, Zenon
          DEFS Phase1b, Send
        <5>5. \A aa, v, c: VotedForIn(aa, v, c) => VotedForIn(aa, v, c)'
          BY <2>4 DEFS VotedForIn
        <5>6. ASSUME NEW c \in m.maxVotedBallot  +  1 .. m.ballot  -  1
              PROVE \neg (\E v \in Values : VotedForIn(m.acc, v, c)')
          <6>1. \neg (\E v \in Values : VotedForIn(m.acc, v, c))
            BY <5>2
          <6>2. SUFFICES ASSUME NEW v \in Values PROVE ~ VotedForIn(m.acc, v, c)'
            BY Zenon
          <6>3. c # mm.ballot
            <7>1. m.maxVotedBallot \in Nat \cup {-1}
              BY <4>2, <5>1, TypeOK DEF TypeOK, Messages
            <7>2. c < m.ballot /\ c \in Int
              BY IntIntervalDef, NatIsInt, <4>2, <5>3, <7>1, Zenon
            <7>3. m.ballot <= mm.ballot
              BY <2>4, IntGeToLe, <3>1, <5>2, <5>3, NatIsInt, IntLeAssoc5
            <7>4. c < mm.ballot
              BY <7>2, <7>3, <5>3, NatIsInt, IntLeAssoc6
            <7> QED
              BY <7>2, <5>3, <7>4, IntGtIsNeq, Zenon
          <6> msg.ballot = mm.ballot
            BY Zenon
          <6> QED
            BY <2>4, <3>1, <4>2, <6>1, <6>3, Zenon DEF VotedForIn
        <5> QED
          BY <3>1, <5>2, <4>1, <5>4, <5>5, <5>6, NatIsInt, Zenon DEFS Phase1b
      <4>3. ASSUME m.type = "2a" PROVE MsgInv!(m)!2!2'
        <5>1. SafeAt(m.val, m.ballot)'
          <6> m.val \in Values /\ m.ballot \in Ballots
            BY TypeOK, <2>4, <3>m, <4>3, Zenon DEF TypeOK, Messages
          <6> QED
            BY <2>0, <2>4, <3>m, <4>3, SafeAtStable, Zenon DEFS Phase1b, MsgInv
        <5>2. \A ma \in msgs': (ma.type = "2a") /\ (ma.ballot = m.ballot)
                                   => (ma = m)
          BY <2>m, <2>4, <3>m, <4>3, Zenon DEFS Phase1b, Send, MsgInv
        <5>3. QED
          BY <5>1, <5>2
      <4>4. ASSUME m.type = "2b" PROVE MsgInv!(m)!3!2'
        <5>1. maxVotedBallot'[m.acc] = mm.ballot
          BY <2>4, <3>1, TypeOK, Zenon DEFS TypeOK
        <5>2. mm.ballot \in Nat
          BY <2>4, TypeOK, Zenon DEFS TypeOK, Messages
        <5>3. maxVotedBallot[a] \in Int /\ maxBallot[a] \in Int
          BY TypeOK, NatIsInt DEFS TypeOK
        <5>4. maxVotedBallot[a] <= mm.ballot
          <6> maxVotedBallot[a] <= maxBallot[a]
            BY <3>1, IntGeToLe, TypeOK DEFS TypeOK
          <6> QED
            BY IntLeAssoc5, NatIsInt, <2>4, <5>2, <5>3
        <5>5. CASE m \in msgs
          <6>1. MsgInv!(m)!3!2
            BY <2>m, <4>4, <5>5 DEF MsgInv
          <6>2. m.ballot <= mm.ballot
            <7>3. m.ballot \in Nat
              BY <5>5, TypeOK DEFS TypeOK, Messages
            <7> QED
              BY <3>1, IntLeAssoc5, NatIsInt, <5>2, <6>1, <5>3, <5>4, <7>3, Zenon
          <6> QED
            BY <5>5, <5>1, <6>2, <6>1, <2>4
        <5>6. CASE m = msg
          <6>1. \E ma \in msgs' : /\ ma.type = "2a"
                                  /\ ma.ballot = m.ballot
                                  /\ ma.val = m.val
            BY <5>6, <2>4, <4>4
          <6>2. m.ballot <= maxVotedBallot'[m.acc]
            BY <5>6, <5>1, IntLeRefl, NatIsInt, <5>2, Zenon
          <6>3. QED
            BY <6>1, <6>2
        <5> QED
          BY <5>5, <5>6, <2>4
      <4>5. QED
        BY <4>2, <4>3, <4>4
    <3>2. CASE m.acc # a
      <4>2. ASSUME m.type = "1b" PROVE MsgInv!(m)!1!2'
        <5>1. m \in msgs
          BY <2>4, <4>2, <3>m, Zenon DEFS TypeOK, Messages
        <5>2. MsgInv!(m)!1!2
          BY <2>m, <4>2, <5>1, Zenon DEFS MsgInv
        <5>3. m.ballot \in Nat /\ mm.ballot \in Nat /\ maxBallot[a] \in Nat \cup {-1}
          BY <5>1, TypeOK, Zenon DEFS TypeOK, Messages
        <5>a. maxBallot'[m.acc] = maxBallot[m.acc]
          BY <2>4, <3>2, <4>2, <5>1, TypeOK, Zenon DEFS TypeOK, Messages
        <5>4. m.ballot <= maxBallot'[m.acc]
          BY <2>4, <3>2, <5>2, <5>a, <5>3, IntLeAssoc5, IntGeToLe, NatIsInt, Zenon
          DEFS Phase1b, Send
        <5>5. \A aa, v, c: VotedForIn(aa, v, c) => VotedForIn(aa, v, c)'
          BY <2>4 DEFS VotedForIn, Send
        <5>6. ASSUME NEW c \in m.maxVotedBallot  +  1 .. m.ballot  -  1
              PROVE \neg (\E v \in Values : VotedForIn(m.acc, v, c)')
          <6> msg.acc # m.acc
            BY <3>2
          <6> QED
            BY <2>4, <3>2, <5>2, Zenon DEF VotedForIn
        <5> QED
          BY <3>2, <5>2, <5>a, <5>4, <5>5, <5>6, NatIsInt, Zenon DEFS Phase1b
      <4>3. ASSUME m.type = "2a" PROVE MsgInv!(m)!2!2'
        <5>1. SafeAt(m.val, m.ballot)'
          <6> m.val \in Values /\ m.ballot \in Ballots
            BY TypeOK, <2>4, <3>m, <4>3, Zenon DEF TypeOK, Messages
          <6> QED
            BY <2>0, <2>4, <3>m, <4>3, SafeAtStable, Zenon DEFS Phase1b, MsgInv
        <5>2. \A ma \in msgs': (ma.type = "2a") /\ (ma.ballot = m.ballot)
                                   => (ma = m)
          BY <2>m, <2>4, <3>m, <4>3, Zenon DEFS Phase1b, Send, MsgInv
        <5>3. QED
          BY <5>1, <5>2
      <4>4. ASSUME m.type = "2b" PROVE MsgInv!(m)!3!2'
        <5> m.acc \in Acceptors
          <6>1. CASE m \in msgs
            BY <4>4, <6>1, TypeOK, Zenon DEFS TypeOK, Messages
          <6>2. CASE m = msg
            BY <4>4, <6>2, TypeOK, Zenon DEFS TypeOK, Messages
          <6> QED
            BY <6>1, <6>2, <2>4
        <5>1. maxVotedBallot'[m.acc] = maxVotedBallot[m.acc]
          BY <2>4, <3>2, TypeOK, Zenon DEFS TypeOK
        <5>2. mm.ballot \in Nat
          BY <2>4, TypeOK, Zenon DEFS TypeOK, Messages
        <5>3. maxVotedBallot[a] \in Int /\ maxBallot[a] \in Int
          BY TypeOK, NatIsInt DEFS TypeOK
        <5>4. maxVotedBallot[a] <= mm.ballot
          <6> maxVotedBallot[a] <= maxBallot[a]
            BY <3>2, IntGeToLe, TypeOK DEFS TypeOK
          <6> QED
            BY IntLeAssoc5, NatIsInt, <2>4, <5>2, <5>3
        <5>5. CASE m \in msgs
          <6>1. MsgInv!(m)!3!2
            BY <2>m, <4>4, <5>5 DEF MsgInv
          <6> QED
            BY <5>5, <5>1, <6>1, <2>4
        <5>6. CASE m = msg
          BY <5>6, <3>2
        <5> QED
          BY <5>5, <5>6, <2>4
      <4>5. QED
        BY <4>2, <4>3, <4>4
    <3>4. QED
      BY <3>1, <3>2
  <2>6. QED
    BY <2>0, <2>1, <2>2, <2>3, <2>4 DEF Next, Phase1b, Phase2b
<1>3. QED

(***************************************************************************)
(* Consistent                                                              *)
(***************************************************************************)
THEOREM Consistent == Spec => []Consistency
<1> USE DEF Ballots
<1>1. Inv => Consistency
  <2> SUFFICES ASSUME Inv,
                      NEW v1 \in Values,  NEW v2 \in Values,
                      NEW b1 \in Ballots, ChosenIn(v1,b1),
                      NEW b2 \in Ballots, ChosenIn(v2,b2)
               PROVE  v1 = v2
    BY DEF Consistency, Chosen
  <2>1. CASE b1 = b2
    <3>1. PICK a1 \in Acceptors : VotedForIn(a1, v1, b1)
      BY QuorumAssumption DEF ChosenIn
    <3>2. PICK a2 \in Acceptors : VotedForIn(a2, v2, b2)
      BY QuorumAssumption DEF ChosenIn
    <3>. QED
      BY <3>1, <3>2, <2>1, VotedOnce DEF Inv
  <2>n. \A x,y \in Nat: x < y => x \in 0..y-1
    BY ONLY SimpleArith
  <2>2. CASE b1 < b2
    <3>1. b1 \in 0 .. b2 - 1
      BY <2>2, <2>n
    <3>2. PICK Q1 \in Quorums: \A a \in Q1: VotedForIn(a, v1, b1)
      BY DEF ChosenIn
    <3>3. PICK Q2 \in Quorums: \A a \in Q2: VotedForIn(a, v2, b1) \/ WontVoteIn(a, b1)
      <4>1. SafeAt(v2, b2)
        BY VotedInv, QuorumNonEmpty, QuorumAssumption DEF ChosenIn, Inv
      <4>2. QED
        BY <2>2, <3>1, <4>1 DEF SafeAt
    <3>4. QED
      <4>1. \A a1 \in Q1, a2 \in Q2:
              VotedForIn(a1, v1, b1) /\ VotedForIn(a2, v2, b1) => v1 = v2
        BY QuorumAssumption, VotedOnce, Zenon DEF Inv
      <4>2. PICK a3 \in Q1 \cap Q2: VotedForIn(a3, v2, b1) \/ WontVoteIn(a3, b1)
        BY QuorumAssumption, <3>3, <3>2, <4>1
      <4>3. QED
        BY <3>2, <4>1, <4>2, Zenon DEF WontVoteIn
  <2>3. CASE b2 < b1
    <3>1. b2 \in 0 .. b1 - 1
      BY <2>3, <2>n
    <3>2. PICK Q1 \in Quorums: \A a \in Q1: VotedForIn(a, v2, b2)
      BY DEF ChosenIn
    <3>3. PICK Q2 \in Quorums: \A a \in Q2: VotedForIn(a, v1, b2) \/ WontVoteIn(a, b2)
      <4>1. SafeAt(v1, b1)
        BY VotedInv, QuorumNonEmpty, QuorumAssumption DEF ChosenIn, Inv
      <4>2. QED
        BY <2>3, <3>1, <4>1 DEF SafeAt
    <3>4. QED
      <4>1. \A a1 \in Q1, a2 \in Q2:
              VotedForIn(a1, v1, b2) /\ VotedForIn(a2, v2, b2) => v1 = v2
        BY QuorumAssumption, VotedOnce, Zenon DEF Inv
      <4>2. PICK a3 \in Q1 \cap Q2: VotedForIn(a3, v1, b2) \/ WontVoteIn(a3, b2)
        BY QuorumAssumption, <3>3, <3>2, <4>1
      <4>3. QED
        BY <3>2, <4>1, <4>2, Zenon DEF WontVoteIn
  <2>4. QED
    <3> \A x,y \in Nat: x = y \/ x < y \/ y < x
      BY ONLY SimpleArith
    <3> QED
      BY <2>1, <2>2, <2>3
<1>2. QED

-----------------------------------------------------------------------------
chosenBar == {v \in Values : Chosen(v)}

C == INSTANCE Consensus WITH chosen <- chosenBar

(***************************************************************************)
(* The following theorem asserts that this specification of Paxos refines  *)
(* the trivial specification of consensus in module Consensus.             *)
(***************************************************************************)
THEOREM Refinement == Spec => C!Spec


=============================================================================
\* Modification History
\* Last modified Mon May 08 04:00:01 EDT 2017 by hernanv
\* Last modified Sun May 07 03:33:37 EDT 2017 by hernanv
\* Last modified Fri Nov 28 10:39:17 PST 2014 by lamport
\* Last modified Sun Nov 23 14:45:09 PST 2014 by lamport
\* Last modified Sat Nov 22 12:04:19 CET 2014 by merz
\* Last modified Fri Nov 21 17:40:41 PST 2014 by lamport
\* Last modified Tue Mar 18 11:37:57 CET 2014 by doligez
\* Last modified Sat Nov 24 18:53:09 GMT-03:00 2012 by merz
\* Created Sat Nov 17 16:02:06 PST 2012 by lamport
